#!/usr/bin/env /opt/opensvc/bin/python
""" 
Verify file content. 
The variable format is json-serialized:

{
  "path": "/some/path/to/file",
  "check": "some pattern into the file",
  "com": "comment to correct manualy"
}
"""

import os
import sys
import json
import stat
import re
import urllib
import tempfile
import codecs

sys.path.append(os.path.dirname(__file__))

from comp import *

MAXSZ = 8*1024*1024

class FileIncCom(object):
    def __init__(self, prefix='OSVC_COMP_FILEINCCOM_'):
        self.prefix = prefix.upper()
        self.files = {}
        self.ok = {}
        self.checks = []
        self.upds = {}

        self.sysname, self.nodename, x, x, self.machine = os.uname()

        for k in [ key for key in os.environ if key.startswith(self.prefix)]:
            try:
                self.add_file(os.environ[k])
            except ValueError:
                print >>sys.stderr, 'key syntax error on var[', k, '] = ',os.environ[k]

        if len(self.checks) == 0:
            raise NotApplicable()

    def fixable(self):
        return RET_NA

    def fix(self):
        return RET_NA

    def get_env_item(self, d, item):
        if item not in d:
            return d
        if type(d[item]) != str and type(d[item]) != unicode:
            return d
        if not d[item].startswith("%%"):
            return d
        s = d[item].strip('%').replace('ENV:', '')
        if not s.startswith('OSVC_COMP_'):
            s = 'OSVC_COMP_'+s
        if s not in os.environ:
            print >>sys.stderr, "%s is not set"%s
            raise
        d[item] = os.environ[s]
        try:
            d[item] = int(d[item])
        except:
            pass
        return d

    def subst(self, v):
        if type(v) == list:
            l = []
            for _v in v:
                l.append(self.subst(_v))
            return l
        if type(v) != str and type(v) != unicode:
            return v
        p = re.compile('%%ENV:\w+%%')
        for m in p.findall(v):
            s = m.strip("%").replace('ENV:', '')
            if s in os.environ:
                _v = os.environ[s]
            elif 'OSVC_COMP_'+s in os.environ:
                _v = os.environ['OSVC_COMP_'+s]
            else:
                print >>sys.stderr, s, 'is not an env variable'
                raise NotApplicable()
            v = v.replace(m, _v)
        return v

    def read_file(self, path):
        if not os.path.exists(path):
            return ''
        out = ''
        try :
            f = codecs.open(path, 'r', encoding="utf8", errors="ignore")
            out = f.read()
            f.close()
        except IOError as (errno, strerror):
            print "cannot read '%s', error=%d - %s" %(path, errno, strerror)
            raise
        except:
            print >>sys.stderr, "Cannot open '%s', unexpected error: %s"%(path, sys.exc_info()[0])
            raise
        return out

    def add_file(self,v):
        r = RET_OK
        d = json.loads(v)
        if 'path' not in d:
            print >>sys.stderr, "'path' should be defined:", d
            r |= RET_ERR
        try:
            d = self.get_env_item(d, 'check')
        except:
            r |= RET_ERR
        for k in ('path', 'check'):
            if k in d:
                d[k] = self.subst(d[k])
        if 'path' in d:
            d['path'] = d['path'].strip()
        if not d['path'] in self.upds:
            self.upds[d['path']] = 0
        if not d['path'] in self.files:
            try:
                fsz = os.path.getsize(d['path'])
            except:
                fsz = 0
            if fsz > MAXSZ:
                self.ok[d['path']] = 0
                self.files[d['path']] = ''
                print >>sys.stderr, "file '%s' is too large [%.2f Mb] to fit" %(d['path'], fsz/(1024.*1024))
                r |= RET_ERR
            else:
                try:
                    self.files[d['path']] = self.read_file(d['path'])
                    self.ok[d['path']] = 1
                except:
                    self.files[d['path']] = ""
                    self.ok[d['path']] = 0
                    r |= RET_ERR
        val = True
        self.checks.append({'check':d['check'], 'path':d['path'], 'valid':val, 'com':d['com']})
        return r
        
    def check(self):
        r = RET_OK
        for ck in self.checks:
            if not ck['valid']:
                print >>sys.stderr, "rule error: '%s' does not match target content" % ck['check']
                r |= RET_ERR
                continue
            if self.ok[ck['path']] != 1:
                r |= RET_ERR
                continue
            pr = RET_OK
            m = 0
            ok = 0
            lines = self.files[ck['path']].split('\n')
            for line in lines:
                if re.match(ck['check'], line):
                    print "line '%s' found in '%s'" %(line, ck['path'])
                    ok += 1
            if ok == 0:
                print >>sys.stderr, "pattern '%s' not found in %s. To correct: %s"%(ck['check'], ck['path'], ck['com'])
                pr |= RET_ERR
            r |= pr
        return r



if __name__ == "__main__":
    syntax = """syntax:
      %s PREFIX check|fixable|fix"""%sys.argv[0]
    if len(sys.argv) != 3:
        print >>sys.stderr, "wrong number of arguments"
        print >>sys.stderr, syntax
        sys.exit(RET_ERR)
    try:
        o = FileInc(sys.argv[1])
        if sys.argv[2] == 'check':
            RET = o.check()
        elif sys.argv[2] == 'fix':
            RET = o.fix()
        elif sys.argv[2] == 'fixable':
            RET = o.fixable()
        else:
            print >>sys.stderr, "unsupported argument '%s'"%sys.argv[2]
            print >>sys.stderr, syntax
            RET = RET_ERR
    except ComplianceError:
        sys.exit(RET_ERR)
    except NotApplicable:
        sys.exit(RET_NA)
    except:
        import traceback
        traceback.print_exc()
        sys.exit(RET_ERR)

    sys.exit(RET)

